@@ -24,7 +24,6 @@ libvirt-based hypervisor.
 """
 
 import configparser
-import crypt
 import datetime
 import fcntl
 import getpass
@@ -76,6 +75,16 @@ def logerr(line):
     if line:
         log.error(line)
 
+def valid_name(name):
+    """
+    Returns true only if name contains a restricted set of characters.
+    """
+    safe = set(string.ascii_letters + string.digits + '-_.')
+    for c in name:
+        if c not in safe:
+            return False
+    return True
+
 # The sh debug logging swamps the log file, so back off.
 def _adjust_sh_log():
     logger = logging.getLogger(sh.__name__)
@@ -130,6 +139,7 @@ class Settings:
         self.arch = get('arch', '')
         self.pool = get('pool', 'default')
         self.user = get('user', get('username', getpass.getuser()))
+        self.password_length = int(get('password-length', 24))
         self.image_format = get('image-format', 'qcow2')
         self.memory = get('memory', 512)
         self.vcpus = get('vcpus', 1)
@@ -211,11 +221,13 @@ class Connection:
     def __enter__(self, uri=None):
         if uri is None:
             uri = libvirt_uri
+        log.debug(f"Opening libvirt connection: uri='{uri}'")
         self.conn = libvirt.open(uri)
         Connection.opens += 1
         return self.conn
 
     def __exit__(self, *exc):
+        log.debug("Closing libvirt connection")
         self.conn.close()
         Connection.closes += 1
 
@@ -232,7 +244,8 @@ class Creds:
         self.password = password
         self.ssh_identity = ssh_identity
 
-    def generate_password(self, length=24):
+    @classmethod
+    def generate_password(cls, length=24):
         """
         Generate a random password consisting of letters and digits.
         """
@@ -244,7 +257,7 @@ class Creds:
         """
         Generate a ssh key pair for passwordless ssh login.
         """
-        ssh_identity = f'{virtup_data_home}/sshkeys/{name}'
+        ssh_identity = f'{virtup_data_home}/sshkeys/{name}/id_rsa'
         if os.path.exists(ssh_identity):
             log.debug(f"SSH key file '{ssh_identity}' already exists.")
             if not os.path.exists(f'{ssh_identity}.pub'):
@@ -547,7 +560,9 @@ class Instance:
                 if line.startswith('IP address'):
                     continue
                 address, _, _, mac, _, _ = line.split()
-                arp[mac] = address
+                if not mac in arp:
+                    arp[mac] = []
+                arp[mac].append(address)
         return arp
 
     def _ping(self, address):
@@ -566,18 +581,18 @@ class Instance:
         """
         Attempt to retreive the instance address from the arp cache.
         """
-        address = None
         mac = self.mac()
         for retries in range(120, -1, -1):
             arp = self._arp_table()
-            address = arp.get(mac)
-            if address and self._ping(address):
-                break
+            addresses = arp.get(mac, [])
+            for address in addresses:
+                if self._ping(address):
+                    return address
             if retries > 0:
                 suffix = 'ies' if retries > 1 else 'y'
                 log.debug(f"Waiting for instance '{self.name}' address in arp cache; {retries} retr{suffix} left.")
                 time.sleep(2)
-        return address
+        return None
 
     def _address_from_dns(self):
         """
@@ -689,9 +704,9 @@ class Instance:
 
     @classmethod
     def build(cls,
-              name,
-              template=None, # defaults to <name>
-              prefix='',
+              template,
+              target=None,
+              prefix='VIRTUP-',
               settings=None,
               root_password=None,
               user=None,
@@ -703,15 +718,21 @@ class Instance:
               dns_domain=None,
               **kwargs):
         """
-        Build an instance with virt-builder and virt-install.
+        Build a base instance with virt-builder and virt-install.
+        Returns the base instance if it already exists.
         """
         _adjust_sh_log()
-        if name is None:
-            raise ValueError('<name> is required.')
-        if not template:
-            template = name
 
-        name = f'{prefix}{name}'  # Optional instance name prefix.
+        if target:
+            name = target
+        else:
+            # template -> base instance name
+            safe = set(string.ascii_letters + string.digits + '_-.')
+            suffix = ''.join([c if c in safe else '-' for c in template])
+            name = f"{prefix}{suffix}"
+
+        if not valid_name(name):
+            raise ValueError(f"Base instance name '{name}' is not valid.")
 
         if cls.exists(name):
             log.info(f"Instance '{name}' already exists.")
@@ -719,6 +740,7 @@ class Instance:
 
         if settings is None:
             settings = Settings(template)
+
         maddrs = MacAddresses()
         path = query_storage_pool(settings.pool)
         if not os.access(path, os.R_OK | os.W_OK):
@@ -739,13 +761,13 @@ class Instance:
         # Generate the ssh keys. Generate the passwords if not provided.
         if not user:
             user = settings.user
+        if not root_password:
+            root_password = Creds.generate_password(settings.password_length)
+        if not password:
+            password = Creds.generate_password(settings.password_length)
         root_creds = Creds('root', password=root_password)
         user_creds = Creds(user, password=password)
 
-        # useradd requires a hashed password.
-        salt = crypt.mksalt(crypt.METHOD_SHA512)
-        password = crypt.crypt(user_creds.password, salt)
-
         # Setup virt-builder arguments.
         if memory is None:
             memory = settings.memory
@@ -770,12 +792,16 @@ class Instance:
                 '--output', image,
                 '--format', settings.image_format,
                 '--hostname', hostname,
-                '--root-password', f'password:{root_creds.password}',
                 '--run-command', 'ssh-keygen -A',
-                '--run-command', f"useradd -m -s /bin/bash -p '{password}' {user}",
-                '--ssh-inject', f'{user}:file:{user_creds.ssh_identity}.pub',
+                '--root-password', f'password:{root_creds.password}',
+                '--ssh-inject', f'{root_creds.username}:file:{root_creds.ssh_identity}.pub',
+                '--copy-in', f"{root_creds.ssh_identity}:/root/.ssh",
+                '--run-command', f"useradd -m -s /bin/bash {user_creds.username}",
+                '--password', f"{user_creds.username}:password:{user_creds.password}",
+                '--ssh-inject', f'{user_creds.username}:file:{user_creds.ssh_identity}.pub',
+                '--copy-in', f"{user_creds.ssh_identity}:/home/{user_creds.username}/.ssh",
                 '--run-command', 'mkdir -p /etc/sudoers.d',
-                '--write',  f'/etc/sudoers.d/99-virt-up:{user} ALL=(ALL) NOPASSWD: ALL',
+                '--write',  f'/etc/sudoers.d/99-virt-up:{user_creds.username} ALL=(ALL) NOPASSWD: ALL',
                 *extra_args)
 
         # Setup virt-install options. Reuse the last mac address for this
@@ -815,6 +841,17 @@ class Instance:
             'root': vars(root_creds),
             'user': vars(user_creds),
             'address-source': settings.address_source,
+            'ssh_options': {
+                'CheckHostIP': 'no',
+                'ControlMaster': 'auto',
+                'ControlPersist': '60s',
+                'ForwardX11': 'no',
+                'IdentitiesOnly': 'yes',
+                'LogLevel': 'ERROR',
+                'PasswordAuthentication': 'no',
+                'StrictHostKeyChecking': 'no',
+                'UserKnownHostsFile': '/dev/null',
+            },
         }
         if size:
             meta['size'] = size
@@ -830,6 +867,9 @@ class Instance:
     def clone(self,
             target,
             settings=None,
+            user=None,
+            password=None,
+            root_password=None,
             hostname=None,
             memory=None,
             size=None,
@@ -846,6 +886,9 @@ class Instance:
         """
         _adjust_sh_log()
         assert(target)
+        if not valid_name(target):
+            raise ValueError(f"target '{target}' contains invalid characters.")
+
         if self.exists(target):
             log.info(f"Target instance '{target}' already exists.")
             return Instance(target)
@@ -865,6 +908,21 @@ class Instance:
         if not os.access(path, os.R_OK | os.W_OK):
             raise PermissionError(f"Read and write access is required for path '{path}'.")
 
+        # Get default values from the current template settings.
+        if not memory:
+            memory = settings.memory
+        if not vcpus:
+            vcpus = settings.vcpus
+        if not graphics:
+            graphics = settings.graphics
+        if dns_domain is None:
+            dns_domain = settings.dns_domain
+        if not hostname:
+            if dns_domain:
+                hostname = f'{target}.{dns_domain}'
+            else:
+                hostname = target
+
         # Clone the image.
         source_image = self.meta['disk']
         target_image = f'{path}/{target}.{settings.image_format}'
@@ -880,15 +938,27 @@ class Instance:
                 extra_args = settings.cp_args
                 cp(*extra_args, source_image, target_image)
 
-        # Setup virt-sysprep args.
-        if not hostname:
-            if dns_domain is None:
-                dns_domain = settings.dns_domain
-            if dns_domain:
-                hostname = f'{target}.{dns_domain}'
-            else:
-                hostname = target
+        # Setup credentials for new instance.
+        if not root_password:
+            root_password = Creds.generate_password(settings.password_length)
+        root_creds = Creds('root', password=root_password)
+        if not user:
+            user = settings.user
+        if not password:
+            password = Creds.generate_password(settings.password_length)
+        user_creds = Creds(user, password=password)
 
+        # Args to setup user creds in cloned instance.
+        user_args = []
+        if user_creds.username != self.meta['user']['username']:
+            user_args.extend(['--run-command', f"useradd -m -s /bin/bash {user_creds.username}"])
+        user_args.extend([
+            '--password', f"{user_creds.username}:password:{user_creds.password}",
+            '--ssh-inject', f'{user_creds.username}:file:{user_creds.ssh_identity}.pub',
+            '--copy-in', f"{user_creds.ssh_identity}:/home/{user_creds.username}/.ssh"
+        ])
+
+        # Setup virt-sysprep args.
         extra_args = settings.virt_sysprep_args
 
         with LockFile():
@@ -898,17 +968,12 @@ class Instance:
                 '--add', target_image,
                 '--operations', 'defaults,-ssh-userdir',
                 '--hostname', hostname,
+                '--root-password', f"password:{root_creds.password}",
+                *user_args,
                 *extra_args)
 
         # Setup virt-install options. Reuse the last mac address for this
         # instance so it will (hopefully) be assigned the same address.
-        if not memory:
-            memory = self.meta.get('memory', 512)
-        if not vcpus:
-            vcpus = self.meta.get('vcpus', 1)
-        if not graphics:
-            graphics = self.meta.get('graphics', 'none')
-
         optional_args = []
         mac = maddrs.lookup(target)
         if mac:
@@ -943,6 +1008,8 @@ class Instance:
         meta['memory'] = memory
         meta['vcpus'] = vcpus
         meta['graphics'] = graphics
+        meta['root'] = vars(root_creds)
+        meta['user'] = vars(user_creds)
         instance = Instance(target, meta=meta)
         maddrs.update(target, instance.mac())
         instance.address() # Wait for an address to be assigned.
@@ -959,16 +1026,6 @@ class Instance:
         Create an ansible inventory file for the cloned instances.
         """
         filename = f'{virtup_data_home}/inventory.yaml'
-        ssh_options = [
-            ('ControlMaster', 'auto'),
-            ('ControlPersist', '60s'),
-            ('ForwardX11', 'no'),
-            ('IdentitiesOnly', 'yes'),
-            ('LogLevel', 'ERROR'),
-            ('StrictHostKeyChecking', 'no'),
-            ('UserKnownHostsFile', '/dev/null'),
-        ]
-        ssh_common_args = ' '.join(['-o %s=%s' % x for x in ssh_options])
         clones = {}
         templates = {}
         for instance in Instance.all():
@@ -983,7 +1040,7 @@ class Instance:
                 'ansible_port': '22',
                 'ansible_private_key_file': instance.meta['user']['ssh_identity'],
                 'ansible_connection': 'ssh',
-                'ansible_ssh_common_args': ssh_common_args,
+                'ansible_ssh_common_args': ' '.join(instance._ssh_option_args()),
             }
             if instance.is_clone():
                 clones[name] = host
@@ -1008,6 +1065,17 @@ class Instance:
                 for key, value in templates[name].items():
                     fp.write(f'          {key}: "{value}"\n')
 
+    def _ssh_option_args(self):
+        """
+        Get the list of ssh option arguments.
+        """
+        args = []
+        options = self.meta.get('ssh_options', {})
+        for k, v in options.items():
+            args.append('-o')
+            args.append(f'{k}={v}')
+        return args
+
     def login(self, mode='ssh', command=None):
         """
         ssh or stfp login to the instance.
@@ -1022,11 +1090,7 @@ class Instance:
         ssh_identity = self.meta['user']['ssh_identity']
         args = [
             '-i', ssh_identity,
-            '-o', 'PasswordAuthentication=no',
-            '-o', 'CheckHostIP=no',
-            '-o', 'UserKnownHostsFile=/dev/null',
-            '-o', 'StrictHostKeyChecking=no',
-            '-o', 'LogLevel=ERROR',
+            *self._ssh_option_args(),
             f'{user}@{address}',
         ]
         if command:
@@ -1051,11 +1115,7 @@ class Instance:
         command = shlex.join(args)
         ssh_args = [
             '-i', ssh_identity,
-            '-o', 'PasswordAuthentication=no',
-            '-o', 'CheckHostIP=no',
-            '-o', 'UserKnownHostsFile=/dev/null',
-            '-o', 'StrictHostKeyChecking=no',
-            '-o', 'LogLevel=ERROR',
+            *self._ssh_option_args(),
             f'{user}@{address}',
             command,
         ]
